"""
@brief   Module to summarise a video into N frames. There are a lot of methods
         for video summarisation, and a lot of repositories in GitHub, but
         none of them seems to work out of the box. This module contains a 
         simple way of doing it.
@author  Luis C. Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date    18 May 2022.
"""

import argparse
import imageio_ffmpeg 
import numpy as np
import cv2
import tqdm
import math
import sklearn_extra.cluster
import skimage.measure
import scipy

# My imports
import videosum


class VideoSummariser():
    def __init__(self, algo, number_of_frames: int = 100, 
            width: int = 1920, height: int = 1080):
        """
        @brief   Summarise a video into a collage.
        @details The frames in the collage are evenly taken from the video. 
                 No fancy stuff.

        @param[in]  algo              Algorithm to select key frames.
        @param[in]  number_of_frames  Number of frames of the video you want
                                      to see in the collage.
        @param[in]  width             Width of the summary collage.
        @param[in]  height            Height of the summary collage.
        """
        assert(algo in VideoSummariser.ALGOS)
        assert(number_of_frames > 0)
        assert(width > 0)
        assert(height > 0)

        self.algo = algo
        self.number_of_frames = number_of_frames
        self.width = width
        self.height = height
        self.form_factor = float(self.width) / self.height

        # Compute the width and height of each collage tile
        self.tile_height = self.height
        nframes = VideoSummariser._how_many_rectangles_fit(self.tile_height, 
                                                          self.width, 
                                                          self.height)
        while nframes < self.number_of_frames: 
            self.tile_height -= 1
            nframes = VideoSummariser._how_many_rectangles_fit(self.tile_height, 
                                                               self.width, 
                                                               self.height)
        self.tile_width = int(round(self.tile_height * self.form_factor))

        # Compute how many tiles per row and column
        self.tiles_per_row = self.width // self.tile_width
        self.tiles_per_col = self.height // self.tile_height 
 
    @staticmethod
    def _how_many_rectangles_fit(tile_height, width, height):
        """
        @brief Given a certain tile height, this method computes how many
               of them fit in a collage of a given size. We assume that the
               form factor of the tiles in the collage has to be the same of 
               the collage itself.
        @param[in]  tile_height  Tile height.
        @param[in]  width        Width of the collage.
        @param[in]  height       Height of the collage.
        """
        # Compute form factor
        ff = float(width) / height

        # Compute rectangle height 
        tile_width = int(round(tile_height * ff))

        # Compute how many rectangles fit inside the collage
        tiles_per_row = width // tile_width 
        tiles_per_col = height // tile_height

        return tiles_per_row * tiles_per_col 


    def _insert_frame(self, im, i, j):
        """@brief Insert image into the collage."""

        # Resize frame to the right tile size
        im_resized = cv2.resize(im, (self.tile_width, self.tile_height), 
                                interpolation = cv2.INTER_LANCZOS4)

        # Insert image within the collage
        y_start = i * self.tile_height
        y_end = y_start + im_resized.shape[0] 
        x_start = j * self.tile_width
        x_end = x_start + im_resized.shape[1] 
        self.collage[y_start:y_end, x_start:x_end] = im_resized

    def get_key_frames_time(self, input_path):
        """
        @brief Get a list of key frames from the video. The key frames are
               simply evenly spaced along the video.
        
        @param[in]  input_path  Path to the video file.

        @returns a list of Numpy/OpenCV BGR images.
        """
        key_frames = []

        # Collect the collage frames from the video 
        reader = imageio_ffmpeg.read_frames(input_path, pix_fmt='rgb24')
        meta = reader.__next__()
        w, h = meta['size']
        nframes = int(math.floor(meta['duration'] * meta['fps']))
        interval = nframes // self.number_of_frames
        counter = interval
        for raw_frame in tqdm.tqdm(reader):
            # If we have collected all the frames we needed, mic out
            if len(key_frames) == self.number_of_frames:
                break
            
            # If this frame is a key frame...
            counter -= 1
            if counter == 0:
                counter = interval

                # Convert video frame into a BGR OpenCV/Numpy image
                im = np.frombuffer(raw_frame, 
                        dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()
            
                # Insert video frame in our list of key frames
                key_frames.append(im)

        return key_frames
        
    def get_key_frames_fid(self, input_path):
        """
        @brief Get a list of key video frames. 
        @details They key frames are selected by unsupervised clustering of 
                 latent feature vectors corresponding to the video frames.
                 The latent feature vector is obtained from Inception V3
                 trained on ImageNet. The clustering method used is kmedoids.  

        @param[in]  input_path  Path to the video file.
        
        @returns a list of Numpy/BGR images, range [0.0, 1.0], dtype = np.uint8. 
        """
        latent_vectors = []

        # Initialise Inception network model
        fid = videosum.FrechetInceptionDistance('vector')

        # Collect feature vectors for all the frames
        reader = imageio_ffmpeg.read_frames(input_path, pix_fmt='rgb24')
        meta = reader.__next__()
        w, h = meta['size']
        for raw_frame in tqdm.tqdm(reader):
            # Convert video frame into a BGR OpenCV/Numpy image
            im = np.frombuffer(raw_frame, dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

            # Compute latent feature vector for this video frame
            vec = fid.get_latent_feature_vector(im)

            # Add feature vector to our list
            latent_vectors.append(vec)

        # Compute mean and variance of every vector
        X = np.array(latent_vectors)

        # Cluster the feature vectors using the Frechet Inception Distance 
        kmedoids = sklearn_extra.cluster.KMedoids(n_clusters=self.number_of_frames, 
            init='k-medoids++',
            random_state=0).fit(X)
        indices = kmedoids.medoid_indices_.tolist()

        # Retrieve the video frames corresponding to the cluster means
        key_frames = []
        counter = -1
        reader = imageio_ffmpeg.read_frames(input_path, pix_fmt='rgb24')
        for raw_frame in tqdm.tqdm(reader):
            counter += 1
            if counter in indices:
                # Convert video frame into a BGR OpenCV/Numpy image
                im = np.frombuffer(raw_frame, dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

                # Add key frame to list
                key_frames.append(im)

        return key_frames

    def get_key_frames_scda(self, input_path):
        """
        @brief Get a list of key video frames. 
        @details They key frames are selected by unsupervised clustering of 
                 latent feature vectors corresponding to the video frames.
                 The latent feature vector is obtained as explained in 
                 Wei et al. 2017:

                 "Selective Convolutional Descriptor Aggregation for
                  Fine-Grained Image Retrieval"
                 
                 The only difference with the paper is we use InceptionV3
                 to retrieve the latent tensor (as opposed to VGG-16).

        @param[in]  input_path  Path to the video file.
        
        @returns a list of Numpy/BGR images, range [0.0, 1.0], dtype = np.uint8. 
    
        """
        latent_vectors = []

        # Initialise Inception network model
        fid = videosum.FrechetInceptionDistance('tensor')

        # Collect feature vectors for all the frames
        reader = imageio_ffmpeg.read_frames(input_path, pix_fmt='rgb24')
        meta = reader.__next__()
        w, h = meta['size']
        for raw_frame in tqdm.tqdm(reader):
            # Convert video frame into a BGR OpenCV/Numpy image
            im = np.frombuffer(raw_frame, dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

            # Compute latent feature vector for this video frame
            tensor = fid.get_latent_feature_tensor(im)

            # Sum tensor over channels
            aggregation_map = tensor.sum(axis=2)

            # Compute mask map
            mean = aggregation_map.mean()
            mask_map = np.zeros_like(aggregation_map)
            mask_map[aggregation_map > mean] = 1

            # Connected component analysis
            cc = skimage.measure.label(mask_map) 

            # Largest connected component
            largest = np.argmax(np.bincount(cc.flat)[1:]) + 1
            largest_cc = scipy.ndimage.binary_fill_holes(cc == largest).astype(cc.dtype)

            # Selected descriptor set
            selected_descriptor_set = []
            for i in range(8):
                for j in range(8):
                    if largest_cc[i, j]:
                        selected_descriptor_set.append(tensor[i, j])
            selected_descriptor_set = np.array(selected_descriptor_set)

            # Compute avg&maxPool SCDA feature vector
            avgpool = np.mean(selected_descriptor_set, axis=0)
            maxpool = np.max(selected_descriptor_set, axis=0)
            scda = np.hstack([avgpool, maxpool])

            # Add feature vector to our list
            latent_vectors.append(scda)

        # Compute mean and variance of every vector
        X = np.array(latent_vectors)

        # Cluster the feature vectors using the Frechet Inception Distance 
        kmedoids = sklearn_extra.cluster.KMedoids(n_clusters=self.number_of_frames, 
            init='k-medoids++',
            random_state=0).fit(X)
        indices = kmedoids.medoid_indices_.tolist()

        # Retrieve the video frames corresponding to the cluster means
        key_frames = []
        counter = -1
        reader = imageio_ffmpeg.read_frames(input_path, pix_fmt='rgb24')
        for raw_frame in tqdm.tqdm(reader):
            counter += 1
            if counter in indices:
                # Convert video frame into a BGR OpenCV/Numpy image
                im = np.frombuffer(raw_frame, dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

                # Add key frame to list
                key_frames.append(im)

        return key_frames

    def get_key_frames(self, input_path):
        return VideoSummariser.ALGOS[self.algo](self, input_path)

    def summarise(self, input_path):
        """
        @brief Create a collage of the video.  

        @param[in]  input_path   Path to an input video.

        @returns a BGR image (numpy.ndarray) with a collage of the video.
        """
        # Get list of key frames
        key_frames = VideoSummariser.ALGOS[self.algo](self, input_path)

        #assert(len(key_frames) == self.number_of_frames)
        if len(key_frames) < self.number_of_frames:
            print('[WARN] The key frame selection algorithm selected ' \
                + 'less frames than you wanted.')

        # Reset collage image
        self.collage = np.zeros((self.tiles_per_col * self.tile_height, 
                                 self.tiles_per_row * self.tile_width, 3),
                                dtype=np.uint8)
        
        # Initialise collage counters
        i = 0
        j = 0
        for im in key_frames:
                # Insert image in the collage
                self._insert_frame(im, i, j) 

                # Update collage iterators
                j += 1
                if j == self.tiles_per_row:
                    j = 0
                    i += 1

        # Put dividing lines on the collage
        #for x in range(1, self.tiles_per_row):
        #    self.collage[:, x * self.tile_width] = 255

        return self.collage
    
    # Class attribute: supported key frame selection algorithms
    ALGOS = {
        'time': get_key_frames_time,
        'fid' : get_key_frames_fid,
        'scda': get_key_frames_scda,
    }


if __name__ == '__main__':
    raise RuntimeError('[ERROR] This is not a Python script.')



"""
@brief   This module provides a class to easily compute the Frechet Inception 
         Distance between to Numpy/OpenCV BGR images.

@author  Luis C. Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date    21 Jul 2022.
"""

import numpy as np
import torch
import scipy

# My imports
import videosum


class FrechetInceptionDistance():
    def __init__(self, mode: str, device: str = 'cuda'):
        self.mode = mode
        self.device = device
        
        # Initialise model
        if self.mode == 'vector':
            self.model = videosum.InceptionV3().to(device)
        elif self.mode == 'tensor':
            self.model = videosum.InceptionV3(output_vector=False).to(device)
        else:
            raise ValueError('[ERROR] Output mode not recognized.')
        self.model.eval()
    
    def _get_latent_features(self, im):
        """
        @brief  Given an image, this method computes a 1D feature vector.

        @param[in]  im  Numpy/OpenCV BGR image, shape (H, W, 3),
                        dtype = np.uint8.
        
        @returns a Numpy ndarray of 2048 channels.
        """
        # Convert the image to RGB in range [0, 1]
        im = im[...,::-1].copy().astype(np.float32) / 255.

        # Convert image from Numpy to Pytorch and upload it to GPU
        im = torch.from_numpy(im).to(self.device)

        # Convert image to shape (1, 3, H, W)
        im = im.permute(2, 0, 1).unsqueeze(0)

        # Run inference on the image and get a vector of shape (2048,)
        with torch.no_grad():
            output = self.model(im)[0].flatten().cpu().detach().numpy()

        return output
    
    def get_latent_feature_vector(self, im):
        assert(self.mode == 'vector')
        return self._get_latent_features(im)
    
    def get_latent_feature_tensor(self, im):
        assert(self.mode == 'tensor')
        return self._get_latent_features(im).reshape((8, 8, 2048))

        
    
    @staticmethod
    def frechet_inception_distance(mu_sigma_1, mu_sigma_2):
        """
        @brief Compute the Frechet inception distance between two 
               distributions.

        @param[in]  mu_sigma_1  Numpy array of arrays [[means], [covs]] for 
                                distribution 1.
        @param[in]  mu_sigma_2  Numpy array of arrays [[means], [covs]] for 
                                distribution 2.
        """
        return FrechetInceptionDistance._calculate_frechet_distance(
            mu_sigma_1[0], mu_sigma_1[1], mu_sigma_2[0], mu_sigma_2[0])

    @staticmethod 
    def _calculate_frechet_distance(mu1, sigma1, mu2, sigma2, eps=1e-6):
        """
        @brief Numpy implementation of the Frechet Distance.
               The code of this method was taken from the pytorch-fid package.
        
        @details The Frechet distance between two multivariate Gaussians 
                 X_1 ~ N(mu_1, C_1) and X_2 ~ N(mu_2, C_2) is:

                 d^2 = ||mu_1 - mu_2||^2 + Tr(C_1 + C_2 - 2*sqrt(C_1*C_2))

        Stable version by Dougal J. Sutherland.

        @param[in]  mu1     Numpy array containing the activations of a layer 
                            of the inception net for generated samples.
        @param[in]  mu2     The sample mean over activations, precalculated 
                            on an representative data set.
        @param[in]  sigma1  The covariance matrix over activations for 
                            generated samples.
        @param[in]  sigma2  The covariance matrix over activations, precalculated 
                            on an representative data set.
    
        @returns the Frechet distance.
        """
        mu1 = np.atleast_1d(mu1)
        mu2 = np.atleast_1d(mu2)

        sigma1 = np.atleast_2d(sigma1)
        sigma2 = np.atleast_2d(sigma2)

        assert mu1.shape == mu2.shape, \
            'Training and test mean vectors have different lengths'
        assert sigma1.shape == sigma2.shape, \
            'Training and test covariances have different dimensions'

        diff = mu1 - mu2

        # Product might be almost singular
        covmean, _ = scipy.linalg.sqrtm(sigma1.dot(sigma2), disp=False)
        if not np.isfinite(covmean).all():
            msg = ('fid calculation produces singular product; '
                   'adding %s to diagonal of cov estimates') % eps
            print(msg)
            offset = np.eye(sigma1.shape[0]) * eps
            covmean = linalg.sqrtm((sigma1 + offset).dot(sigma2 + offset))

        # Numerical error might give slight imaginary component
        if np.iscomplexobj(covmean):
            if not np.allclose(np.diagonal(covmean).imag, 0, atol=1e-3):
                m = np.max(np.abs(covmean.imag))
                raise ValueError('Imaginary component {}'.format(m))
            covmean = covmean.real

        tr_covmean = np.trace(covmean)

        return (diff.dot(diff) + np.trace(sigma1)
                + np.trace(sigma2) - 2 * tr_covmean) 


if __name__ == '__main__':
    raise RuntimeError('[ERROR] The FID module is not a script.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import setuptools
import unittest

# Read the contents of the README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(name='videosum',
    version='0.0.3',
    description='Python module to summarise a video into a collage.',
    author='Luis C. Garcia-Peraza Herrera',
    author_email='luiscarlos.gph@gmail.com',
    license='MIT License',
    url='https://github.com/luiscarlosgph/endoseg',
    packages=['videosum'],
    package_dir={'videosum' : 'src'}, 
    install_requires = [
        'numpy', 
        'opencv-python',
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    #test_suite = 'tests',
)



"""
@brief  Unit tests to check the videosum package.
@author Luis Carlos Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date   7 June 2022.
"""
import unittest
import cv2
import numpy as np

# My imports
import videosum

class TestVideosum(unittest.TestCase):

    def test_collage(self):
        """
        @brief Simple test to check that the collage still works with
               newer versions of the dependencies.
        """
        # Load video
        width = 640
        height = 480
        nframes = 9
        vs = videosum.VideoSummariser(nframes, width, height)

        # Make collage
        new_collage = vs.summarise('test/data/test.mp4')

        # Compare the collage with the one stored
        old_collage = cv2.imread('test/data/test.png')

        self.assertTrue(np.allclose(new_collage, old_collage))


if __name__ == '__main__':
    unittest.main()



"""
@brief   Module to summarise a video into N frames. There are a lot of methods
         for video summarisation, and a lot of repositories in GitHub, but
         none of them seems to work out of the box. This module contains a 
         simple way of doing it.
@author  Luis C. Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date    18 May 2022.
"""

import argparse
import imageio_ffmpeg 
import numpy as np
import os
import cv2
import tqdm

# My imports
import videosum


def help(short_option):
    """
    @returns The string with the help information for each command line option.
    """
    help_msg = {
        '-i': 'Path to the video (required: True)',
        '-o': 'Output directory path (required: True)',
        '-n': 'Number of frames in the collage (required: True)',
        '-x': 'Width of the collage (required: True)',
        '-y': 'Height of the collage (required: True)',
        '-a': 'Algorithm for frame selection, available options: time, frechet (required: True)',
    }
    return help_msg[short_option]


def parse_cmdline_params():
    """@returns The argparse args object."""
    # Create command line parser
    parser = argparse.ArgumentParser(description='Upload LEMON dataset to Synapse.')
    parser.add_argument('-i', '--input', required=True, type=str, 
                        help=help('-i'))
    parser.add_argument('-o', '--output', required=True, type=str, 
                        help=help('-o'))
    parser.add_argument('-n', '--nframes', required=True, type=int, 
                        help=help('-n'))
    parser.add_argument('-x', '--width', required=True, type=int, 
                        help=help('-x'))
    parser.add_argument('-y', '--height', required=True, type=int, 
                        help=help('-y'))
    parser.add_argument('-a', '--algo', required=True, type=str,
                        help=help('-a'))

    # Read parameters
    args = parser.parse_args()
    
    return args


def validate_cmdline_params(args):
    """
    @brief Input directory must exist and output must not.
    """
    if not os.path.isfile(args.input):
        raise RuntimeError('[ERROR] Input file does not exist.')
    if os.path.isfile(args.output):
        raise RuntimeError('[ERROR] Output file already exists.')
    return args


def main():
    # Read command line parameters
    args = parse_cmdline_params()
    validate_cmdline_params(args)

    vidsum = videosum.VideoSummariser(args.algo, args.nframes, args.width, args.height)
    im = vidsum.summarise(args.input)
    cv2.imwrite(args.output, im)


if __name__ == '__main__':
    main()
